ORIGINAL_METHOD_PREFIX = "original_"
PATCHED_LOG_METHOD_PREFIX = "log_"
PANDAS_ADDITIONAL_METHODS_TO_OVERIDE = ["copy", "reset_index"]
PANDAS_METHODS_TO_OVERIDE = [
    "query",
    "drop",
    "dropna",
    "assign",
    "sort_index",
    "sort_values",
    "head",
    "tail",
    "sample",
    "fillna",
    "merge",
    "join",
    "nlargest",
    "nsmallest",
    "apply",
    "applymap",
    "pipe",
    "rolling",
    "groupby",
    "rename",
    "agg",
    "aggregate",
    "stack",
    "unstack",
    "pivot",
    "pivot_table",
    "mask",
    "max",
    "mean",
    "median",
    "melt",
    "replace",
    "skew",
    "notna",
    "kurt",
    "expanding",
    "drop_duplicates",
    "bfill",
    "corr",
    "corrwith",
    "droplevel",
    "explode",
    "ffill",
    "filter",
    "first",
    "kurtosis",
    "align",
    "transform",
    "update",
    "squeeze",
    "shift",
    "rank",
    "nunique",
    "min",
    "mod",
    "mode",
    "std",
]



import pandas as pd
from pandas_log import enable


df = pd.read_csv("../examples/pokemon.csv")

with enable(True):
    df.query("type_1=='Fire' or type_2=='Fire'").drop(
        "type_1", axis=1
    ).nlargest(1, "total")
#     # Wont fix df2 = df.copy()
#
#
# wont write to stdout df.head()
df.query("legendary==0")
# FIX 1: df.copy().query("legendary==0")
# Probably all objects that are df need to be updated



import itertools
from contextlib import suppress
from functools import partial, wraps
from inspect import signature

import pandas as pd

import pandas_flavor as pf
from pandas_log import patched_logs_functions, settings
from pandas_log.timer import Timer


def set_df_attr(df, attr_name, attr_value):
    """ Hacky way to set attributes in dataframe

        :param df: DataFrame
        :param attr_name: Attribute name
        :param attr_value: Attribute value
        :return: None
    """

    df.__dict__[attr_name] = attr_value


def append_df_attr(df, attr_name, attr_value):
    """ Hacky way to append a value to dataframe

        :param df: DataFrame
        :param attr_name: Attribute name
        :param attr_value: Attribute value
        :return: None
    """

    df.__dict__[attr_name].append(attr_value)


def get_df_attr(df, attr_name, default_val):
    """ Get Dataframe attribute if exists otherwise default value

        :return: Dataframe attribute
    """

    return df.__dict__.get(attr_name, default_val)


def get_pandas_func(func, prefix=settings.ORIGINAL_METHOD_PREFIX):
    """ Get original pandas method

        :param func: pandas method name
        :param prefix: the prefix used to keep original method
        :return: Original pandas method
    """

    return getattr(pd.DataFrame, f"{prefix}{func.__name__}")


def get_signature_repr(fn, args, full_signature=True):
    """ Get the signature for the original pandas method with actual values

        :param fn: The pandas method
        :param args: The arguments used when it was applied
        :return: string representation of the signature for the applied pandas method
    """

    def _get_bold_text(text):
        return f"\033[1m{text}\033[0m"

    def _get_orig_func_params():
        return [
            param_value if full_signature else param_name
            for param_name, param_value in signature(
                get_pandas_func(fn)
            ).parameters.items()
            if param_name not in ("kwargs", "self")
        ]

    def _get_param_value(param_with_default, arg_value):
        res = str(param_with_default)
        if arg_value is not None:
            param_name = res.split("=")[0]
            arg_value = (
                f'"{arg_value}"' if isinstance(arg_value, str) else arg_value
            )
            res = (
                param_name
                if isinstance(arg_value, pd.DataFrame)
                else f"{param_name}={arg_value}"
            )
        return res

    zip_func = itertools.zip_longest if full_signature else zip
    orig_func_params = _get_orig_func_params()
    args_vals = ", ".join(
        _get_param_value(param_with_default, arg_value)
        for param_with_default, arg_value in zip_func(orig_func_params, args)
    )
    return f"{_get_bold_text(fn.__name__)}({args_vals}):"


def get_patch_log_func(func, prefix=settings.PATCHED_LOG_METHOD_PREFIX):
    """ Get return relevant log function.
        For example: for .query pandas' method as input one will get ._log_query method
        from patched_functions

        :param func: The pandas method we applied and going to calculate stats
        :return: a function which calculate statistic for the applied pandas method
    """

    log_method = getattr(patched_logs_functions, "log_default")
    with suppress(AttributeError):
        log_method = getattr(
            patched_logs_functions, f"{prefix}{func.__name__}"
        )

    return log_method


def restore_pandas_func_copy(func, prefix=settings.ORIGINAL_METHOD_PREFIX):
    """ Restore the original pandas method instead of overridden one

        :param func: pandas method name
        :param prefix: the prefix used to keep original method
        :return: None
    """

    original_method = getattr(pd.DataFrame, func)
    setattr(pd.DataFrame, func.replace(prefix, ""), original_method)


def keep_pandas_func_copy(func, prefix=settings.ORIGINAL_METHOD_PREFIX):
    """ Saved copy of the pandas method before it overridden

        :param func: pandas method name
        :param prefix: the prefix used to keep original method
        :return: None
    """

    original_method = getattr(pd.DataFrame, func)
    setattr(pd.DataFrame, f"{prefix}{func}", original_method)


def create_overide_pandas_func(func, silent, full_signature):
    """ Create overridden pandas method dynamically with
        additional logging using DataFrameLogger

        Note: if we extracting _overide_dataframe_method outside we need to implement decorator like here
              https://stackoverflow.com/questions/10176226/how-do-i-pass-extra-arguments-to-a-python-decorator

        :param func: pandas method name to be overridden
        :param silent: Whether additional the statistics get printed
        :param full_signature: adding additional information to function signature
        :return: the same function with additional logging capabilities
    """

    def _get_step_stats(
        fn, fn_args, fn_kwargs, input_df, output_df, exec_time, step_number
    ):
        from pandas_log.step_stats import StepStats

        func_logs = _get_logs_for_specifc_method(
            fn, fn_args, fn_kwargs, input_df, output_df
        )

        step_stats = StepStats(
            exec_time, fn, fn_args, full_signature, step_number, func_logs
        )
        return step_stats

    def _persist_execution_stats(input_df, output_df, step_stats, step_number):
        set_df_attr(output_df, "execution_step_number", step_number)
        prev_exec_history = get_df_attr(input_df, "execution_history", [])
        set_df_attr(output_df, "execution_history", prev_exec_history)
        append_df_attr(output_df, "execution_history", step_stats)

    def _get_logs_for_specifc_method(
        fn, fn_args, fn_kwargs, input_df, output_df
    ):
        fn_kwargs["kwargs"] = fn_kwargs.copy()
        log_method = get_patch_log_func(fn)
        log_method = partial(log_method, output_df, input_df)
        return log_method(*fn_args, **fn_kwargs)

    def _overide_dataframe_method(fn):
        @pf.register_dataframe_method
        @wraps(fn)
        def wrapped(*args, **fn_kwargs):
            # Execute original method and save execution time in t
            with Timer() as t:
                output_df = get_pandas_func(fn)(*args, **fn_kwargs)

            # Prepare variables
            input_df, fn_args, exec_time = args[0], args[1:], t.exec_time
            step_number = get_df_attr(input_df, "execution_step_number", 0) + 1

            # calculate steps stats and persist them into the dataframes
            step_stats = _get_step_stats(
                fn,
                fn_args,
                fn_kwargs,
                input_df,
                output_df,
                exec_time,
                step_number,
            )
            _persist_execution_stats(
                input_df, output_df, step_stats, step_number
            )

            if not silent:
                print(step_stats)

            return output_df

        return wrapped

    return exec(
        f"@_overide_dataframe_method\ndef {func}(df, *args, **kwargs): pass"
    )


if __name__ == "__main__":
    pass



import warnings
from time import time

# TODO should fork humanize for cleaner Timer interface or to extract
# TODO this class to a small useful package

with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    import humanize


class Timer:
    @staticmethod
    def get_humanized_exec_time(end, start):
        """ Return the time it took a function to run in a human friendly way

            :param end: time exactly after method was applied
            :param start: time right before method was applied
            :return: execution time in human friendly way
        """

        exec_time = end - start
        exec_time_humanize = humanize.naturaldelta(exec_time)
        if exec_time_humanize == "a moment":
            exec_time_humanize = f"{exec_time} seconds."
        return f"\t* Step Took {exec_time_humanize}"

    def __enter__(self, *args, **kwargs):
        self.start = time()
        return self

    def __exit__(self, *args, **kwargs):
        self.exec_time = Timer.get_humanized_exec_time(
            end=time(), start=self.start
        )


if __name__ == "__main__":
    pass



# -*- coding: utf-8 -*-

"""Top-level package for pandas-log."""
from .pandas_log import *

__author__ = """Eyal Trabelsi"""
__email__ = "eyaltrabelsi@gmail.com"
__version__ = "0.0.0"

ORIGINAL_METHOD_PREFIX = "original_"



from pandas_log.aop_utils import get_signature_repr
from pandas_log.settings import PANDAS_ADDITIONAL_METHODS_TO_OVERIDE


class StepStats:
    def __init__(
        self,
        exec_time,
        fn,
        fn_args,
        full_signature,
        step_number,
        func_logs=None,
    ):
        """ Constructor

            :param exec_time: the time it took a function to run in a human friendly way
            :param fn: The original pandas method
            :param fn_args: The original pandas method args
            :param full_signature: adding additional information to function signature
            :param step_number: number of of operation in sequence
            :param func_logs: additional logs
        """

        self.exec_time = exec_time
        self.func_sig = get_signature_repr(fn, fn_args, full_signature)
        self.func_logs = func_logs
        self.step_number = step_number
        self.fn = fn

    def __repr__(self):
        """ Return string representation of DataFrameLogger aka
            Statistics regarding applied function

            :return: string representation of StepStats
        """

        func_logs = f"\n{self.func_logs}" if self.func_logs else ""
        step_number = (
            "   "
            if self.fn.__name__ in PANDAS_ADDITIONAL_METHODS_TO_OVERIDE
            else f"\n{self.step_number}) "
        )
        return f"\n{step_number}{self.func_sig}{func_logs}\n{self.exec_time}"



import pandas as pd

# Rows messages
REMOVED_NO_ROWS_MSG = "\t* No change in number of rows of input df."
FILTERED_ROWS_MSG = "\t* Removed {rows_removed} rows ({rows_removed_pct}%), {rows_remaining} rows remaining."

# Cols messages
REMOVED_NO_COLS_MSG = "\t* Removed no columns."
FILTERED_COLS_MSG = "\t* Removed the following columns ({cols_removed}) now only have the following columns ({cols_remaining})."
ASSIGN_EXISTING_MSG = "\t* The columns {existing_cols} were reassigned."
ASSIGN_NEW_MSG = "\t* The columns {new_cols} were created."

# N/A messages
FILLNA_NO_NA_MSG = "\t* There are no nulls."
FILLNA_WITHH_NA_MSG = "\t* Filled {} with {}."

# Mege messages
JOIN_ROWS_MSG = "\t* Number of rows changed, after join is {output_rows} rows."
JOIN_NEW_COLS_MSG = "\t* Added {num_new_columns} columns ({new_columns})."
JOIN_TYPE_MSG = "\t* Its a {how} join with the following cardinality:\n\t\t> rows only in left is {left_only}.\n\t\t> rows only in right is {right_only}.\n\t\t> rows in both is {both}."

# Pick messages
SAMPLE_MSG = "\t* Picked random sample of {output_rows} rows."
NLARGEST_MSG = "\t* Picked {n} largest rows by columns ({cols})."
NSMALLEST_MSG = "\t* Picked {n} smallest rows by columns ({cols})."
HEAD_MSG = "\t* Picked the first {} rows."
TAIL_MSG = "\t* Picked the last {} rows."

# Others
SORT_VALUES_MSG = "\t* Sorting by columns {by} in a {'ascending' if ascending else 'descending'} order."
SORT_INDEX_MSG = "\t* Sorting by index in a {'ascending' if ascending else 'descending'} order."
NOT_IMPLEMENTED_MSG = "\t*Log not implemented yet for this function."
DEFAULT_STRATEGY_USED_MSG = (
    "\t* using default strategy (some metric might not be relevant)"
)
TRANSFORMED_TO_DF_MSG = "\t* After transformation we received Series"


def rows_removed(input_df, output_df):
    return len(input_df) - len(output_df)


def rows_removed_pct(input_df, output_df):
    return 100 * (rows_removed(input_df, output_df)) / len(input_df)


def rows_remaining(output_df):
    return len(output_df)


def cols_removed(input_df, output_df):
    return ",".join(set(input_df.columns) - set(output_df.columns))


def cols_remaining(output_df):
    return ",".join(set(output_df.columns))


def is_same_cols(input_df, output_df):
    return len(input_df.columns) == len(output_df.columns)


def columns_changed(df, cols):
    return set(df.columns).intersection(set(cols))


def columns_added(df, cols):
    return set(cols) - set(df.columns)


def is_same_rows(input_df, output_df):
    return len(input_df) == len(output_df)


def num_of_na(df):
    return df.isnull().values.sum()


def str_new_columns(input_df, output_df):
    return ",".join(set(output_df.columns) - set(input_df.columns))


def num_new_columns(input_df, output_df):
    return len(set(output_df.columns) - set(input_df.columns))


def log_default(output_df, input_df, **kwargs):
    logs = [DEFAULT_STRATEGY_USED_MSG]

    if isinstance(output_df, pd.DataFrame):
        if not is_same_cols(input_df, output_df):
            logs.append(
                FILTERED_COLS_MSG.format(
                    cols_removed=cols_removed(input_df, output_df),
                    cols_remaining=cols_remaining(output_df),
                )
            )
        if not is_same_rows(input_df, output_df):
            logs.append(
                FILTERED_ROWS_MSG.format(
                    rows_removed=rows_removed(input_df, output_df),
                    rows_removed_pct=rows_removed_pct(input_df, output_df),
                    rows_remaining=rows_remaining(output_df),
                )
            )
    elif isinstance(output_df, pd.Series):
        logs.append(TRANSFORMED_TO_DF_MSG)
    return "\n".join(logs)


def log_drop(
    output_df,
    input_df,
    labels=None,
    axis=0,
    index=None,
    columns=None,
    level=None,
    inplace=False,
    errors="raise",
    *args,
    **kwargs,
):
    logs = []
    if is_same_cols(input_df, output_df):
        logs.append(REMOVED_NO_COLS_MSG)
    else:
        logs.append(
            FILTERED_COLS_MSG.format(
                cols_removed=cols_removed(input_df, output_df),
                cols_remaining=cols_remaining(output_df),
            )
        )
    if is_same_rows(input_df, output_df):
        logs.append(REMOVED_NO_ROWS_MSG)
    else:
        logs.append(
            FILTERED_ROWS_MSG.format(
                rows_removed=rows_removed(input_df, output_df),
                rows_removed_pct=rows_removed_pct(input_df, output_df),
                rows_remaining=rows_remaining(output_df),
            )
        )
    return "\n".join(logs)


def log_dropna(
    output_df,
    input_df,
    axis=0,
    how="any",
    thresh=None,
    subset=None,
    inplace=False,
    **kwargs,
):
    logs = []
    if is_same_cols(input_df, output_df):
        logs.append(REMOVED_NO_COLS_MSG)
    else:
        logs.append(
            FILTERED_COLS_MSG.format(
                cols_removed=cols_removed(input_df, output_df),
                cols_remaining=cols_remaining(output_df),
            )
        )
    if is_same_rows(input_df, output_df):
        logs.append(REMOVED_NO_ROWS_MSG)
    else:
        logs.append(
            FILTERED_ROWS_MSG.format(
                rows_removed=rows_removed(input_df, output_df),
                rows_removed_pct=rows_removed_pct(input_df, output_df),
                rows_remaining=rows_remaining(output_df),
            )
        )
    return "\n".join(logs)


def log_assign(output_df, input_df, **kwargs):
    logs = []
    cols = kwargs.keys()
    if columns_changed(input_df, cols):
        logs.append(
            ASSIGN_EXISTING_MSG.format(
                existing_cols=",".join(columns_changed(input_df, cols))
            )
        )
    if columns_added(input_df, cols):
        logs.append(
            ASSIGN_NEW_MSG.format(
                new_cols=",".join(columns_added(input_df, cols))
            )
        )
    return "\n".join(logs)


def log_query(output_df, input_df, expr, inplace=False, *args, **kwargs):
    if is_same_rows(input_df, output_df):
        return REMOVED_NO_ROWS_MSG
    else:
        return FILTERED_ROWS_MSG.format(
            rows_removed=rows_removed(input_df, output_df),
            rows_removed_pct=rows_removed_pct(input_df, output_df),
            rows_remaining=rows_remaining(output_df),
        )


def log_reset_index(output_df, input_df, **kwargs):
    return ""


def log_sort_index(
    output_df,
    input_df,
    axis=0,
    level=None,
    ascending=True,
    inplace=False,
    kind="quicksort",
    na_position="last",
    sort_remaining=True,
    by=None,
    **kwargs,
):
    return SORT_INDEX_MSG.format(ascending)


def log_sort_values(
    output_df,
    input_df,
    by,
    axis=0,
    ascending=True,
    inplace=False,
    kind="quicksort",
    na_position="last",
    **kwargs,
):
    return SORT_VALUES_MSG.format(by, ascending)


def log_tail(output_df, input_df, n=5, **kwargs):
    return TAIL_MSG.format(n)


def log_head(output_df, input_df, n=5, **kwargs):
    return HEAD_MSG.format(n)


def log_merge(
    output_df,
    input_df,
    right,
    how="inner",
    on=None,
    left_on=None,
    right_on=None,
    left_index=False,
    right_index=False,
    sort=False,
    suffixes=("_x", "_y"),
    copy=True,
    indicator=False,
    validate=None,
    **kwargs,
):
    logs = []
    merged = input_df.original_merge(
        right,
        "outer",
        on,
        left_on,
        right_on,
        left_index,
        right_index,
        sort,
        suffixes,
        copy,
        True,
        validate,
    )
    logs.append(
        JOIN_TYPE_MSG.format(how=how, **merged._merge.value_counts().to_dict())
    )
    if is_same_rows(input_df, output_df):
        logs.append(REMOVED_NO_ROWS_MSG)
    else:
        logs.append(JOIN_ROWS_MSG.format(output_rows=len(output_df)))
    if not is_same_cols(input_df, output_df):
        logs.append(
            JOIN_NEW_COLS_MSG.format(
                num_new_columns=num_new_columns(input_df, output_df),
                new_columns=str_new_columns(input_df, output_df),
            )
        )

    return "\n".join(logs)


def log_join(
    output_df,
    input_df,
    other,
    on=None,
    how="left",
    lsuffix="",
    rsuffix="",
    sort=False,
    **kwargs,
):
    logs = []
    merged = input_df.original_merge(
        other, "outer", on, input_df.index, other.index, indicator=True
    )
    logs.append(
        JOIN_TYPE_MSG.format(how=how, **merged._merge.value_counts().to_dict())
    )

    if is_same_rows(input_df, output_df):
        logs.append(REMOVED_NO_ROWS_MSG)
    else:
        logs.append(JOIN_ROWS_MSG.format(output_rows=len(output_df)))
    if not is_same_cols(input_df, output_df):
        logs.append(
            JOIN_NEW_COLS_MSG.format(
                num_new_columns=num_new_columns(input_df, output_df),
                new_columns=str_new_columns(input_df, output_df),
            )
        )
    return "\n".join(logs)


def log_fillna(
    output_df,
    input_df,
    value=None,
    method=None,
    axis=None,
    inplace=False,
    limit=None,
    downcast=None,
    **kwargs,
):
    if num_of_na(input_df) == num_of_na(output_df):
        return FILLNA_NO_NA_MSG
    else:
        return FILLNA_WITHH_NA_MSG.format(num_of_na(input_df), value)


def log_sample(
    output_df,
    input_df,
    n=None,
    frac=None,
    replace=False,
    weights=None,
    random_state=None,
    axis=None,
    *args,
    **kwargs,
):
    return SAMPLE_MSG.format(output_rows=len(output_df))


def log_nlargest(output_df, input_df, n, columns, keep="first", **kwargs):
    return NLARGEST_MSG.format(n=n, cols=columns)


def log_nsmallest(output_df, input_df, n, columns, keep="first", **kwargs):
    return NSMALLEST_MSG.format(n=n, cols=columns)


if __name__ == "__main__":
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

requirements = []

setup_requirements = ['pytest-runner', ]

test_requirements = ['pytest>=3',]

setup(
    name='pandas-log',
    version='0.1.2',
    description="pandas-log provides feedback about basic pandas operations. It provides simple wrapper functions for the most common functions, such as apply, map, query and more.",
    author="Eyal Trabelsi",
    author_email='eyaltrabelsi@gmail.com',
    url='https://github.com/eyaltrabelsi/pandas-log',
    packages=find_packages(include=['pandas_log', 'pandas_log.*']),
    install_requires=requirements,
    python_requires=">=3.4",
    license="MIT license",
    long_description=readme,
    long_description_content_type="text/x-rst",
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements
)



# -*- coding: utf-8 -*-

"""Main module."""

import warnings
from contextlib import contextmanager

import pandas as pd

from pandas_log import settings
from pandas_log.aop_utils import (
    create_overide_pandas_func,
    keep_pandas_func_copy,
    restore_pandas_func_copy,
)

__all__ = ["auto_enable", "auto_disable", "enable"]


ALREADY_ENABLED = False


def auto_disable():
    """ Restore original pandas method without the additional log functionality (statistics)
        Note: we keep the original methods using original_ prefix.
        :return: None
    """
    global ALREADY_ENABLED
    if not ALREADY_ENABLED:
        return

    for func in dir(pd.DataFrame):
        if func.startswith(settings.ORIGINAL_METHOD_PREFIX):
            restore_pandas_func_copy(func)
    ALREADY_ENABLED = False


@contextmanager
def enable(verbose=False, silent=False, full_signature=True):
    """ Adds the additional logging functionality (statistics) to pandas methods only for the scope of this
        context manager.

        :param verbose: Whether some inner functions should be recorded as well.
                        For example: when a dataframe being copied
        :param silent: Whether additional the statistics get printed
        :param full_signature: adding additional information to function signature
        :return: None
    """

    auto_enable(verbose, silent, full_signature)
    yield
    auto_disable()


def auto_enable(verbose=False, silent=False, full_signature=True):
    """ Adds the additional logging functionality (statistics) to pandas methods.

        :param verbose: Whether some inner functions should be recorded as well.
                        For example: when a dataframe being copied
        :param silent: Whether additional the statistics get printed
        :param full_signature: adding additional information to function signature
        :return: None
    """
    global ALREADY_ENABLED
    if ALREADY_ENABLED:
        return

    if verbose:
        settings.PANDAS_METHODS_TO_OVERIDE.extend(
            settings.PANDAS_ADDITIONAL_METHODS_TO_OVERIDE
        )

    # Suppressing warning of the fact we override pandas functions.
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        for func in dir(pd.DataFrame):
            if func in settings.PANDAS_METHODS_TO_OVERIDE:
                keep_pandas_func_copy(func)
                create_overide_pandas_func(func, silent, full_signature)
    ALREADY_ENABLED = True


if __name__ == "__main__":
    pass
